;;;; tdtd.el --- Tony's DTD formatting macros
;; $Id: tdtd.el,v 1.11 1998/05/24 18:44:08 tkg Exp $

;; Copyright (C) 1996, 1997 Tony Graham

;; Author: Tony Graham <tgraham@mulberrytech.com>

;;; This file is not part of GNU Emacs.

;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License
;; as published by the Free Software Foundation; either version 2
;; of the License, or (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software
;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


;;;; Commentary:

;; Macros for editing DTDs

;; Requires tdtd-font.el

;; Send bugs to tgraham@mulberrytech.com


;;;; Code:
(provide 'tdtd)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Version information

(defconst tdtd-version "0.5"
  "Version number of tdtd.")

(defun tdtd-version ()
  "Returns the value of the variable tdtd-version."
  tdtd-version)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Variables

(defvar dtd-autodetect-xml t
  "*Non-nil enables autodetection of XML when entering dtd-mode.")

(defvar dtd-xml-flag nil
  "*Non-nil enables XML-specific behaviour, where implemented.")

(defun dtd-xml-p ()
  "Returns t when XML-specific behaviour is enabled, otherwise returns nil."
  (if dtd-xml-flag
      t nil))

(defvar dtd-indent-tabs-mode nil
  "*Initial value of indent-tabs-mode on entering dtd-mode")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Character sequences

(defvar dtd-sgml-mdo "<!"
  "*Markup Delimiter Open (MDO) character sequence")

(defvar dtd-sgml-mdc ">"
  "*Markup Delimiter Close (MDC) character sequence")

(defvar dtd-comment-start "<!--"
  "*Comment start character sequence")

(defvar dtd-comment-end "-->"
  "*Commend end character sequence")

;; SGML Syntactic Literals
(defvar dtd-empty-literal "EMPTY"
  "EMPTY element syntactic literal")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Variables controlling indentation

(defvar dtd-dtd-max-column 70
  "*Rightmost column for text in the DTD")

(defvar dtd-mdc-indent-column (1- dtd-dtd-max-column)
  "*Column at which to insert a MDC")

(defvar dtd-comment-start-column 25
  "*Column for starting text in a comment")

(defvar dtd-comment-max-column (- dtd-dtd-max-column 3)
  "*Maximum column number for text in a comment")

;; (defvar dtd-sgml-comment-max-column (- dtd-dtd-max-column 2))

(defvar dtd-attribute-name-column 16
  "*Indent for inserting attribute names in attribute definitions")

(defvar dtd-element-name-column 12
  "*Column for element name in an element declaration")

(defvar dtd-element-tag-omission-column 25
  "*Column for inserting the omissibility indicators, if used")

(defvar dtd-element-content-spec-start-column 29
  "*Column at which to start content model")

(defvar dtd-element-content-spec-continuation-column 30
  "*Column at which to start second and subsequent lines of content model")

(defvar dtd-line-comment
  (concat
   dtd-comment-start
   " "
   (make-string
    (- dtd-dtd-max-column
       (length (concat dtd-comment-start "  " dtd-comment-end)))
    ?=)
   " "
   dtd-comment-end
   "\n")
  "*Separator comment line")

(defvar dtd-init-comment-column 16)

(defvar dtd-init-comment-fill-prefix
  (concat
   dtd-comment-start
   (make-string (- dtd-init-comment-column (length dtd-comment-start))
		?\ ))
  "*Prefix for initial comment lines")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Functions

;; If you want to untabify everything every time, add this to functions
;;  (untabify (point-min) (point-max))
;; Its easier to just not insert tabs using:
;;  (setq indent-tabs-mode nil)
(defun dtd-untabify-all ()
  "Untabify the entire file"
  (interactive)
  (untabify (point-min) (point-max)))

(defun dtd-fix-entities ()
  "Quick and dirty addition of \";\" to entity references lacking it"
  (interactive)
  (save-excursion
    (goto-char (point-min))
    (while
	(re-search-forward "\\([%&][-A-Za-z0-9.:_]+\\)\\([^-A-Za-z0-9.:_;]\\)" nil t)
      (replace-match "\\1;\\2" nil nil)
      ;; match 2 may be the start of another entity reference
      (goto-char (match-beginning 2)))))

(defun dtd-indent-or-newline-to (target-column)
  "Indent to TARGET COLUMN or, if at or past target, insert newline and indent."
  (if (>= (current-column)
	  target-column)
      (insert "\n"))
  (indent-to target-column))

(defun dtd-center-comment (comment)
  "Center a comment on line"
  (interactive "sComment: ")
  (insert dtd-comment-start)
  (indent-to (/ (- dtd-dtd-max-column (length comment)) 2))
  (insert comment)
  (indent-to dtd-comment-max-column)
  (insert dtd-comment-start)
  (insert "\n"))

(defun dtd-recenter-comment ()
  "Recenter text in a presumably modified comment line"
  (interactive)
  (save-excursion
    (save-match-data
      (beginning-of-line)
      (if (looking-at "^<!--\\s-*\\(\\w+\\(\\W+\\w+\\)*\\)\\s-*-->$")
	  (let ((contents (match-string 1)))
	    (delete-region (match-beginning 0) (match-end 0))
	    (dtd-center-comment contents))))))

;; TDTD house style puts all comments starting on a favourite column
(defun dtd-comment (comment)
  "Insert COMMENT starting at the usual column"
  (interactive "sComment: ")
  (insert "\n")
  (backward-char)
  (let ((fill-column (1- dtd-comment-max-column))
	(fill-prefix (make-string (1- dtd-comment-start-column) ?\ ))
	(comment-start dtd-init-comment-fill-prefix)
	(saved-auto-fill-function auto-fill-function))
    (auto-fill-mode 1)
    (insert dtd-comment-start)
    (indent-to (1- dtd-comment-start-column))
    (fill-region (point) (save-excursion
			   (insert comment)
			   (point))
		 nil
		 1
		 1)
    ;; The fill does the right thing, but it always ends with
    ;; an extra newline, so we delete the newline.
    (delete-backward-char 1)
    (if (not saved-auto-fill-function)
	(auto-fill-mode 0))
    (indent-to dtd-comment-max-column)
    (insert dtd-comment-end)
    (font-lock-fontify-block)))

(defun dtd-big-comment (comment)
  "Insert COMMENT with <!-- === --> comments above and below.

For example:

<!-- ============================================================= -->
<!--                    This is a \"Big Comment\"                    -->
<!-- ============================================================= -->"
  (interactive "sComment: ")
  (insert dtd-line-comment)
  (dtd-comment comment)
  (insert "\n")
  (insert dtd-line-comment))

(defun dtd-recomment ()
  "Fix text position in a presumably modified comment line"
  (interactive)
  (save-excursion
    (save-match-data
      (beginning-of-line)
      (if (looking-at "^\\(<!--\\*\\)?\\s-*\\([^ \t\n>]+\\([- \t]+[^- \t\n>]+\\)*\\)\\s-*\\(\\*-->\\)?$")
	  (let ((contents (match-string 2)))
	    (delete-region (match-beginning 0) (match-end 0))
	    (dtd-comment contents))))))

;; doesn't work yet.
(defun dtd-join-comments ()
  "Join two comment lines.  Doesn't work yet."
  (interactive)
  (save-excursion
    (save-match-data
      (beginning-of-line)
      (if (looking-at "^\\(<!--\\)?\\s-*\\([^ \t\n>]+\\([- \t]+[^- \t\n>]+\\)*\\)\\s-*\\(-->\\)?$")
	  (let ((contents (match-string 2)))
	    (delete-region (match-beginning 0) (match-end 0))
	    (dtd-comment contents))))))

(defun dtd-declare-element (element-name element-comment element-tag content-spec)
  "Insert an element type declaration at the current point.

ELEMENT-NAME is a descriptive name for the element.  It is output as a
comment.

ELEMENT-COMMENT is a further comment about the element.  The comment
is not output if ELEMENT-CONTENT is an empty string.

ELEMENT-TAG is the element type name as it appears in start- and
end-tags.

CONTENT-SPEC is the element type's content specification.

When dtd-xml-flag in nil, the omitted tag minimization parameter is
output as part of the element type declaration.  When CONTENT-SPEC is
\"EMPTY\", the minimization parameter is \"- o\", otherwise it is \"- -\"

An example inserted element type declaration is as follows:

<!--                    Element name                               -->
<!--                    Comment about the element: what it's for,
                        or something about its behaviour           -->
<!ELEMENT  element-tag  - - (insert, your, content, specification,
                             here)                                   >
"
  (interactive "sElement name: \nsElement Comment: \nsElement Tag: \nsContent Spec: ")
  (dtd-comment element-name)
  (insert "\n")
  (if (not
       (string-equal element-comment ""))
      (progn
	(dtd-comment element-comment)
	(insert "\n")))
  (insert "\n")
  (backward-char)
  (insert "<!ELEMENT  ")
  (insert element-tag)
  (if (not dtd-xml-flag)
      (progn
	(dtd-indent-or-newline-to (1- dtd-element-tag-omission-column))
	(if (string-equal content-spec dtd-empty-literal)
	    (insert "- o")
	  (insert "- -"))))
  (dtd-indent-or-newline-to (1- dtd-element-content-spec-start-column))
  (let ((fill-column dtd-mdc-indent-column)
	(fill-prefix "                             ")
	(saved-auto-fill-function auto-fill-function))
    (auto-fill-mode 1)
    (fill-region (point) (save-excursion
			   (insert content-spec)
			   (point))
		 nil
		 1
		 1)
    ;; The fill does the right thing, but it always ends with
    ;; an extra newline, so we delete the newline.
    (delete-backward-char 1)
    (if (not saved-auto-fill-function)
	(auto-fill-mode 0))
    (dtd-indent-or-newline-to dtd-mdc-indent-column)
    (insert (concat dtd-sgml-mdc "\n"))
    (auto-fill-mode nil)
    (font-lock-fontify-block)))

(defun dtd-insert-mdc ()
  "Without moving point, indent to usual column and insert MDC (>).
MDC = Markup Declaration Close"
  (interactive)
  (save-excursion
    (let ((current-point (point))
	  (mdc-column 0)
	  (mdc dtd-sgml-mdc))
      (if (if (looking-at dtd-sgml-mdo)
	      1
	    (re-search-backward (concat "^" dtd-sgml-mdo)))
	  (progn
	    (if (looking-at dtd-comment-start)
	      (progn
		(setq mdc dtd-comment-end)
		(setq mdc-column dtd-comment-max-column))
	      (progn
		(setq mdc dtd-sgml-mdc)
		(setq mdc-column dtd-mdc-indent-column)))
	    (goto-char current-point)
	    (beginning-of-line)
	    (cond
	     ;; The main "looking-at" regular expression doesn't handle
	     ;; line containing just "-->" or ">" very well, so handle
	     ;; as special case.
	     ((looking-at "[ \t]*\\(--\\)?>$")
	      (progn
		;; Delete to the end of line so we can then insert our mdc
	       (delete-region (point) (save-excursion
					(end-of-line)
					(point)))
	       (indent-to mdc-column)
	       (insert mdc)))
	     ;; This is the main "looking-at" regular expression that
	     ;; handles most things
	     ((looking-at "^\\(\\(<!\\|[ \t]*\\)?\\(--\\)?[ \t]*\\([^ \t\n>]+\\([- \t]+[^- \t\n>]+\\)*\\)?\\)[ \t]*\\(\\(--\\)?>?\\)?$")
	      (progn
		;;(message ":%s:%s:%s:%s:" mdc mdc-column (match-string 4) (match-string 1))
		(goto-char (match-end 1))
		(delete-region (point) (save-excursion
					 (end-of-line)
					 (point)))
		(indent-to mdc-column)
		(insert mdc))))
	    (font-lock-fontify-block))))))

;; A work in progress
(defun dtd-fill-paragraph (&optional justify)
  "DTD fill paragraph function.  A work in progress."
  (interactive "P")
  ;;(insert "DTD")
  (save-excursion
    (let ((current-point (point)))
      (if (not (looking-at dtd-sgml-mdo))
	  (re-search-backward (concat "^" dtd-sgml-mdo)))
      (cond
       ((looking-at "\\(<!ELEMENT\\)")
	(goto-char (match-end 1))
	(if (looking-at "[ \t\r\n]+")
	    (delete-region (match-beginning 0) (match-end 0)))
	(dtd-indent-or-newline-to (1- dtd-element-name-column))
	(looking-at "")
	t)
       ;; If we're not doing anything, return nil so the built-in function
       ;; will run
       (t nil)))))

(defun dtd-comment-setup ()
  "Setup the auto-fill variables for comments in the body of the DTD"
  (interactive)
  (setq fill-prefix (make-string (1- dtd-comment-start-column) ?\ ))
  (auto-fill-mode 'true)
  (setq fill-column dtd-comment-max-column))

(defun dtd-design-comment-setup ()
  "Setup the auto-fill variables for \"Design Considerations\" comments"
  (interactive)
  (setq fill-prefix "      ")
  (auto-fill-mode 'true)
  (setq fill-column dtd-comment-max-colum))

(defun dtd-declaration-setup ()
  "Setup the auto-fill variables for declarations"
  (interactive)
  (setq fill-prefix (make-string
		     (1- dtd-element-content-spec-continuation-column)
		     ?\ ))
  (auto-fill-mode 'true)
  (setq fill-column dtd-comment-max-colum))

(defun dtd-init-comment-setup ()
  "Setup the auto-fill variables for initial comments"
  (interactive)
  (setq fill-prefix dtd-init-comment-fill-prefix)
  (auto-fill-mode 'true)
  (setq fill-column (- dtd-dtd-max-column (length dtd-comment-end))))

;;(setq fill-paragraph-function 'dtd-fill-paragraph)

(defun dtd-startup ()
  "Functions to perform when starting with DTD stuff"
  (interactive)
  (setq auto-fill-function 'nil))

;; Not quite sure what this gets me, but these variables seem to apply
;; in auto-fill mode
(make-local-variable 'comment-start)
(setq comment-start "<!-- ")
(make-local-variable 'comment-end)
(setq comment-end " -->")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Tags stuff

;; Aren't shell escapes fun!
;; "\" -> "\\\\\\"
;; any other significant character -> "\\" + character
(defvar dtd-etags-regex-option
  "--regex=/\\<\\!\\\\\\(ELEMENT\\\\\\|ENTITY\\[\\ \\\\\\t]\\+%\\\\\\|NOTATION\\\\\\|ATTLIST\\\\\\)\\[\\ \\\\\\t]\\+\\\\\\(\\[^\\ \\\\\\t]\\+\\\\\\)/\\\\\\2/"
  "*Complete, including \"--regex=\", etags regular exression option string
for the etags command line for extracting tags (in the Emacs sense)
from DTDs.")

(defvar dtd-etags-program "etags"
  "*Name (and possibly path) of the etags program")

(defvar dtd-etags-output-file "TAGS"
  "*Name of the etags output file")

(defun dtd-etags (filespec)
  "Execute etags on FILESPEC and match on DTD-specific regular expressions.

Runs the \"etags\" program on the specified file with a regular
expression for finding element, entity, attribute list, and notation
declarations.  dtd-tags then modifies the output tags file to
disambiguate parameter entity names versus other names.

FILESPEC is the names or regular expression for the files to be
scanned by etags.  dtd-etags uses shell-command to execute etags
specifically so regular expressions and multiple filenames are
supported, and \"*.dtd\" and \"*.dtd *.ent\" are both valid FILESPEC
values.

dtd-etags also uses the dtd-etags-program, dtd-etags-regex-option, and
dtd-etags-output-file variables to construct the command passed to
shell-command."
  (interactive "sFiles: ")
  (shell-command
   (format "%s %s --output=%s %s"
	   dtd-etags-program
	   dtd-etags-regex-option
	   dtd-etags-output-file
	   filespec))
  (save-excursion
    (save-window-excursion
      ;; We could be visiting the TAGS file, which causes an
      ;; unecessary complication when Emacs prompts to see if
      ;; we want to reload the buffer, so we just kill it out
      ;; of hand.  Since more than one buffer can be visiting it,
      ;; we make sure we catch them all.
      (while (get-file-buffer dtd-etags-output-file)
	(kill-buffer (get-file-buffer dtd-etags-output-file)))
      (find-file dtd-etags-output-file)
      (perform-replace
       "^\\(<!ENTITY[ \t]+%[ \t]+[^\177]+\177\\)\\([^%\1]+\\)" "\\1%\\2;"
       nil t nil)
      ;; Don't bother saving a backup
      (setq backup-inhibited t)
      (save-buffer)
      (kill-buffer (current-buffer)))))

(require 'tdtd-font "tdtd-font")


;;;###autoload
(define-derived-mode dtd-mode
  sgml-mode "DTD"
  "Major mode for SGML and XML DTDs.

dtd-mode features include:

 - dtd-etags function for creating Emacs TAGS files for easy lookup of
   any element, parameter entity, or notation's definition using
   Emacs's built-in tag-lookup functions;

 - Font lock highlighting of declarations so that the important
   information stands out;

 - XML-specific behaviour that, at user option, is triggered by
   automatic detection of the XML Declaration; and

 - Functions for writing and editing declarations and comments to ease
   both creating and keeping a consistent style.

dtd-mode builds on sgml-mode, and the full sgml-mode functions are
still available.

dtd-mode uses many user-definable variables to control the formatting
of declarations, some of which are shown in the following examples:

 		        dtd-comment-start-column    dtd-dtd-max-column
                        |       	     dtd-comment-max-column  |
                        |       				  |  |
<!--                    This is a comment                          -->

           dtd-element-name-column
           |    	dtd-element-tag-ommission-column
           |            |   dtd-element-content-spec-start-column
           |            |   |dtd-element-content-spec-continuation-column
           |            |   ||  		    dtd-dtd-max-column
           |            |   ||  				     |
<!ELEMENT  element-tag  - - (insert, your, content, specification,
                             here)                                   >

\\{dtd-mode-map}"
  ;; XML specific behaviour can be specific to a buffer
  (make-local-variable 'dtd-xml-flag)
  ;; Maybe select XML-specific behaviour if we have an XML declaration
  (if dtd-autodetect-xml
      (save-excursion
	(goto-char (point-min))
	(if (looking-at "[ \t\r\n]*<\\?xml")
	    (setq dtd-xml-flag t))))
  ;; dtd font-lock highlighting setup
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults '(dtd-font-lock-keywords t))
  (setq font-lock-mark-block-function 'dtd-font-lock-mark-block-function)
  ;; Maybe insert space characters when user hits "Tab" key
  (setq indent-tabs-mode dtd-indent-tabs-mode))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Mode map stuff

(define-key dtd-mode-map "\C-c>"     'dtd-insert-mdc)
(define-key dtd-mode-map "\C-c\C-e"  'dtd-declare-element)
(define-key dtd-mode-map "\C-cc"     'dtd-comment)
(define-key dtd-mode-map "\C-cb"     'dtd-big-comment)
(define-key dtd-mode-map "\M-g\M-f"  'font-lock-fontify-buffer)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Syntax table stuff

;; set the non-alphanumeric characters in XML names to
;; 'symbol constituent' class
(modify-syntax-entry ?: "_" dtd-mode-syntax-table)
(modify-syntax-entry ?_ "_" dtd-mode-syntax-table)
(modify-syntax-entry ?- "_ 1234" dtd-mode-syntax-table)
(modify-syntax-entry ?. "_" dtd-mode-syntax-table)
;; "-" is a special case because it is the first and second characters
;; of the start- and end-comment sequences.
(modify-syntax-entry ?- "_ 1234" dtd-mode-syntax-table)
;; "%" does double duty in parameter entity declarations and references.
;; Not necessary to make "%" and ";" act like parentheses since the
;; font lock highlighting tells you when you've put the ";" on the
;; end of a parameter entity reference.
(modify-syntax-entry ?% "_" dtd-mode-syntax-table)
(modify-syntax-entry ?\; "_" dtd-mode-syntax-table)
;; "/" is just punctuation in DTDs, and really only has a role in
;; Formal Public Identifiers
(modify-syntax-entry ?/ "." dtd-mode-syntax-table)
;; Sometimes a string is more than just a string, Dr Freud.
;; Unfortunately, the syntax stuff isn't fussy about matching
;; on paired delimeters, and will happily match a single quote
;; with a double quote, and vice versa.  At least the font
;; lock stuff is more fussy and won't change colour if the
;; delimiters aren't paired.
(modify-syntax-entry ?\" "$" dtd-mode-syntax-table)
(modify-syntax-entry ?\' "$" dtd-mode-syntax-table)
;; The occurrence indicators and connectors are punctuation to us.
(modify-syntax-entry ?| "." dtd-mode-syntax-table)
(modify-syntax-entry ?, "." dtd-mode-syntax-table)
(modify-syntax-entry ?& "." dtd-mode-syntax-table)
(modify-syntax-entry ?? "." dtd-mode-syntax-table)
(modify-syntax-entry ?+ "." dtd-mode-syntax-table)
(modify-syntax-entry ?* "." dtd-mode-syntax-table)
;; "#" is syntax too
(modify-syntax-entry ?# "_" dtd-mode-syntax-table)

;;;; Last provisions
;;;(provide 'tdtd)

;;; tdtd.el ends here
